// parse.simple.c
// (C) 1998 lars brinkhoff.

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "safe.h"
#include "parser.h"
#include "objects.h"

Object *objects[100];

struct
{
  char *name;
  Material *pointer;
} materials[10];

Sky sky;

Real sky_day (Sky *sky, Vector *d)
{
  return 0.5 * vector_scalar (sky->up, *d) + 0.5;
}

Real sky_black (Sky *sky, Vector *d)
{
  return 0;
}

static void
check_f (Object *o, Vector *p, Vector *n, Material *m)
{
  static const Material black = (Material){ 0, 0.1, 0, 0.1, 0 };
  static const Material white = (Material){ 0, 0.1, 0.9, 0.1, 0 };

  if (((int)(0.05 * p->x + 1e6) & 1) ^ ((int)(0.05 * p->z + 1e6) & 1))
    *m = black;
  else
    *m = white;
}

Material *
create_material (Real e, Real s, Real d, Real b)
{
  Material *m;
  
  m = safe_malloc (sizeof (Material));
  
  m->emit = e;
  m->spec = s;
  m->diff = d;
  m->grain = b;
  m->f = 0;

  return m;
}

static void
skip_whitespace (void)
{
  int c;

  while ((c = getchar ()) != EOF &&
	 index (" \t\r\n", c) != 0)
    ;
  
  ungetc (c, stdin);
}

static int
parse_int (void)
{
  int c;
  int i = 0;
  int negative = 0;
  
  skip_whitespace ();

  c = getchar ();
  if (c == '-')
    negative = 1;
  else
    ungetc (c, stdin);

  while ((c = getchar ()) != EOF &&
	 index ("0123456789", c) != 0)
    i = 10 * i + c - '0';

  ungetc (c, stdin);
  
  if (negative)
    return -i;
  else
    return i;
}  

static Real
parse_real (void)
{
  Real r;
  int c;

  r = parse_int ();

  c = getchar ();
  if (c == '.')
    {
      Real x = 0.1;
      
      while ((c = getchar ()) != EOF &&
	     index ("0123456789", c) != 0)
	{
	  r += x * (c - '0');
	  x *= 0.1;
	}
    }

  ungetc (c, stdin);

  return r;
}

static void
expect_char (int d)
{
  int c;

  c = getchar ();
  if (c != d)
    {
      fprintf (stderr, "'%c' expected instead of '%c'\n", d, c);
      exit (1);
    }
}

static Vector
parse_vector (void)
{
  Vector v;
  
  skip_whitespace ();
  expect_char ('(');
  v.x = parse_real ();
  skip_whitespace ();
  expect_char (',');
  v.y = parse_real ();
  skip_whitespace ();
  expect_char (',');
  v.z = parse_real ();
  skip_whitespace ();
  expect_char (')');

  return v;
}

static char *
parse_identifier (void)
{
  char id[100], *id_;
  int i = 0;
  int c;

  skip_whitespace ();
  
  if ((c = getchar ()) != EOF &&
      index ("_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ", c) != 0)
    {
      id[i++] = c;
      
      while ((c = getchar ()) != EOF &&
	     index ("_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
		    "0123456789", c) != 0)
	if (i < 98)
	  id[i++] = c;
    }
  id[i] = 0;
  
  ungetc (c, stdin);
  
  id_ = strdup (id);
  if (id_ == 0)
    {
      fprintf (stderr, "out of memory\n");
      exit (1);
    }

  return id_;
}

static Material *
parse_material (void)
{
  char *n;
  int i;

  n = parse_identifier ();

  for (i = 0; materials[i].pointer != 0; i++)
    if (strcmp (n, materials[i].name) == 0)
      {
	free (n);
	return materials[i].pointer;
      }
  
  free (n);
  return 0;
}

#include <fcntl.h>

static void
parse_input (void)
{
  int i = 0;
  int j = 0;
  int c;

#if 1
    sky.f = 0;
#else
    sky.f = sky_day;
    sky.up = vec (0, 1, 0);
#endif

  objects[0] = 0;
  materials[0].pointer = 0;

  while (!feof (stdin))
    {
      skip_whitespace ();
      c = getchar ();

      switch (c)
	{
	case 's':
	  {
	    Material *m;
	    Vector p;
	    Real r;
	    
	    p = parse_vector ();
	    r = parse_real ();
	    m = parse_material ();

	    objects[i++] = create_sphere (p, r, m, &E);
	  }
	break;
	case 'r':
	  {
	    Material *m;
	    Vector p, n;
	    Real r1, r2;
	    
	    p = parse_vector ();
	    n = parse_vector ();
	    r1 = parse_real ();
	    r2 = parse_real ();
	    m = parse_material ();

	    objects[i++] = create_ring (p, n, r1, r2, m, &E);
	  }
	break;
	case 'M':
	  {
	    Real e, s, d, b;
	    Material *m;
	    char *n;
	    
	    n = parse_identifier ();
	    e = parse_real ();
	    s = parse_real ();
	    d = parse_real ();
	    b = parse_real ();
	    
	    m = create_material (e, s, d, b);

	    materials[j].name = n;
	    materials[j++].pointer = m;

	    materials[j].pointer = 0;
	  }
	break;
	case 'R':
	  {
	    int width, height, rpp;

	    width = parse_int ();
	    height = parse_int ();
	    rpp = parse_int ();
	    
	    render (width, height, rpp);
	    
	    return;
	  }
	break;
	case 'A':
	  {
	    int width, height, x, y, dx, dy, rpp;

	    x = parse_int ();
	    y = parse_int ();
	    dx = parse_int ();
	    dy = parse_int ();
	    width = parse_int ();
	    height = parse_int ();
	    rpp = parse_int ();
	    
	    render_area (x, y, dx, dy, width, height, rpp);
	  }
	break;
	default:
	  fprintf (stderr, "unrecognized entity: '%c'\n", c);
	  exit (1);
	}
      objects[i] = 0;
    }
}

Parser *
create_simple_parser (void)
{
  Parser *parser = safe_malloc (sizeof (Parser));

  parser->read_scene = parse_input;
  //parser->read_commands = read_commands;

  return parser;
}
