/*

ray.h

(C) 1997, 1998 lars brinkhoff

*/

#ifndef RAY_H
#define RAY_H

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

typedef double Real;

static const Real epsilon = 1e-100;
static const Real infinity = 1e100;
static const Real pi = M_PI;

static inline Real sqr (Real x)
{
  return x * x;
}

typedef struct
{
  Real x, y, z;
} Vector;

typedef struct
{
  Vector p, d;
  Real t, wavelength;
  int depth;
} Ray;

typedef struct
{
  Real x[4][4];
} Matrix;

#define vec(x, y, z) ((Vector){ x, y, z})  

static inline Real
vector_scalar (Vector u, Vector v)
{
  return u.x * v.x + u.y * v.y + u.z * v.z;
}

static inline Real
vector_magnitude (Vector v)
{
  return (Real)sqrt (vector_scalar (v, v));
}

#define vector_normalize(v) \
({ \
  Real magnitude = vector_magnitude (v); \
  Real c; \
  \
  if (magnitude == 0.0) \
    fprintf (stderr, "Normalizing zero vector!\n"); \
  else \
    { \
      c = 1 / magnitude; \
      (v).x *= c; \
      (v).y *= c; \
      (v).z *= c; \
    } \
})

static inline Real myrand (Real a, Real b, int n)
{
  int i;
  Real sum = 0;

  for (i = 0; i < n; i++)
    sum += (Real)random () / RAND_MAX;

  return (b - a) * (sum / n) + a;
}

static const Matrix E = (Matrix) { { { 1, 0, 0, 0 },
                                     { 0, 1, 0, 0 },
                                     { 0, 0, 1, 0 },
                                     { 0, 0, 0, 1 } } };

static inline Matrix
matrix_rot_x (Real a)
{
  Matrix m = E;
  m.x[1][1] = cos (a); m.x[1][2] = -sin (a);
  m.x[2][1] = sin (a); m.x[2][2] = cos (a);
  return m;
}

static inline Matrix
matrix_rot_y (Real a)
{
  Matrix m = E;
  m.x[0][0] = cos (a); m.x[0][2] = -sin (a);
  m.x[2][0] = sin (a); m.x[2][2] = cos (a);
  return m;
}

static inline Matrix
matrix_rot_z (Real a)
{
  Matrix m = E;
  m.x[0][0] = cos (a); m.x[0][1] = -sin (a);
  m.x[1][0] = sin (a); m.x[1][1] = cos (a);
  return m;
}

static inline Matrix
matrix_scale (Real x, Real y, Real z)
{
  Matrix m = E;
  m.x[0][0] = x;
  m.x[1][1] = y;
  m.x[2][2] = z;
  return m;
}

static inline Matrix
matrix_translate (Vector delta)
{
  Matrix m = E;
  m.x[0][3] = delta.x;
  m.x[1][3] = delta.y;
  m.x[2][3] = delta.z;
  return m;
}

static inline Matrix
matrix_multiply (Matrix m1, Matrix m2)
{
  int i, j, k;
  Matrix m;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      for (k = 0, m.x[i][j] = 0; k < 4; k++)
	m.x[i][j] += m1.x[i][k] * m2.x[k][j];
  
  return m;
}

static inline Matrix
matrix_add (Matrix A, Matrix B)
{
  int i, j;
  Matrix C;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      C.x[i][j] = A.x[i][j] + B.x[i][j];

  return C;
}

static inline Matrix
matrix_scalar (Real a, Matrix m)
{
  Matrix M;
  int i, j;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      M.x[i][j] = a * m.x[i][j];

  return M;
}

static inline void
matrix_ptr_interpolate_with_E (Matrix *A, Real t, const Matrix *B)
  {
    Real s = 1 - t;

    A->x[0][0] = s + t * B->x[0][0];
    A->x[0][1] =     t * B->x[0][1];
    A->x[0][2] =     t * B->x[0][2];
    A->x[0][3] =     t * B->x[0][3];
    A->x[1][0] =     t * B->x[1][0];
    A->x[1][1] = s + t * B->x[1][1];
    A->x[1][2] =     t * B->x[1][2];
    A->x[1][3] =     t * B->x[1][3];
    A->x[2][0] =     t * B->x[2][0];
    A->x[2][1] =     t * B->x[2][1];
    A->x[2][2] = s + t * B->x[2][2];
    A->x[2][3] =     t * B->x[2][3];
    A->x[3][0] =     t * B->x[3][0];
    A->x[3][1] =     t * B->x[3][1];
    A->x[3][2] =     t * B->x[3][2];
    A->x[3][3] = s + t * B->x[3][3];
  }

static inline Vector
vector_transform (Matrix m, Vector u)
{
  Real U[4], V[4];
  int i, j;
  Vector v;

  U[0] = u.x; U[1] = u.y; U[2] = u.z; U[3] = 1;
  
  for (i = 0; i < 4; i++)
    for (j = 0, V[i] = 0; j < 4; j++)
      V[i] += m.x[i][j] * U[j];

  v.x = V[0]; v.y = V[1]; v.z = V[2];

  return v;
}

static inline void
vector_ptr_transform (Vector *a, Matrix *A, Vector *b)
{
  a->x = A->x[0][0] * b->x + A->x[0][1] * b->y +
         A->x[0][2] * b->z + A->x[0][3];
  a->y = A->x[1][0] * b->x + A->x[1][1] * b->y +
         A->x[1][2] * b->z + A->x[1][3];
  a->z = A->x[2][0] * b->x + A->x[2][1] * b->y +
         A->x[2][2] * b->z + A->x[2][3];
}

static inline void
vector_ptr_transform_no_translation (Vector *a, Matrix *A, Vector *b)
{
  a->x = A->x[0][0] * b->x + A->x[0][1] * b->y +
         A->x[0][2] * b->z;
  a->y = A->x[1][0] * b->x + A->x[1][1] * b->y +
         A->x[1][2] * b->z;
  a->z = A->x[2][0] * b->x + A->x[2][1] * b->y +
         A->x[2][2] * b->z;
}

#define xv(m, x, y, z) vector_transform (m, vec (x, y, z))

#endif
